/**********
 * Copyright (c) 2003-2004 Greg Parker.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY GREG PARKER ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **********/

#include "includes.h"
#include "about.h"
#include "formutils.h"
#include "data/connectionlist.h"
#include "rsrc/rsrc.h"
#include "ssh/ssh.h"
#include <PalmChars.h>

#include "terminalform.h"


typedef enum {
    topAll = -1, // initial state - everything visible
    topVT100 = 0,
    topABC, 
    top123, 
    topFkey,
    topEtc 
} top_pane_t;

typedef enum {
    bottomAll = -1, // initial state - everything visible
    bottomTC = 0,
    bottomFull, 
    bottomCust
} bottom_pane_t;

typedef enum {
    rightNone = -1, // initial state - everything visible
    rightScroll = 0,
    rightArrows
} right_pane_t;

typedef enum {
    caseUpper,
    caseLower, 
    caseCaps
} case_t;

static top_pane_t TopPane = -1;
static bottom_pane_t BottomPane = -1;
static right_pane_t RightPane = -1;


extern struct ssh_session_t *ss;
extern UInt16 NetLibCount;

static Boolean OpenNetLib(void) FORMS_SEGMENT;
static Boolean StartConnection(void) FORMS_SEGMENT;
static void ConstructTopVT100Pane(void) FORMS_SEGMENT;
static void ConstructTopABCPane(void) FORMS_SEGMENT;
static void DestructTopVT100Pane(void) FORMS_SEGMENT;
static void DestructTopABCPane(void) FORMS_SEGMENT;
static void SetABCCase(case_t newcase) FORMS_SEGMENT;
static void ConstructTop123Pane(void) FORMS_SEGMENT;
static void DestructTop123Pane(void) FORMS_SEGMENT;
static void ConstructTopPane(top_pane_t newPane) FORMS_SEGMENT;
static void DestructTopPane(top_pane_t newPane) FORMS_SEGMENT;
static void SetTopPane(top_pane_t newPane) FORMS_SEGMENT;
static void ConstructBottomTCPane(void) FORMS_SEGMENT;
static void DestructBottomTCPane(void) FORMS_SEGMENT;
static void ConstructBottomFullPane(void) FORMS_SEGMENT;
static void DestructBottomFullPane(void) FORMS_SEGMENT;
static void ConstructBottomPane(bottom_pane_t newPane) FORMS_SEGMENT;
static void DestructBottomPane(bottom_pane_t newPane) FORMS_SEGMENT;
static void SetBottomPane(bottom_pane_t newPane) FORMS_SEGMENT;
static void ShowABCPanes(void) FORMS_SEGMENT;
static void Show123Panes(void) FORMS_SEGMENT;
static void HideABC123Panes(void) FORMS_SEGMENT;
static char control_character(char c) FORMS_SEGMENT;
static void SendModifiedKey(struct ssh_session_t *ss, unsigned char c, int ctrlPressed, int metaPressed) FORMS_SEGMENT;
static void ClearModifiers(void) FORMS_SEGMENT;
static void SendModifiedKeyAndClearModifiers(unsigned char c) FORMS_SEGMENT;
static void SendUpArrowKey(void) FORMS_SEGMENT;
static void SendDownArrowKey(void) FORMS_SEGMENT;
static void SendRightArrowKey(void) FORMS_SEGMENT;
static void SendLeftArrowKey(void) FORMS_SEGMENT;
static Boolean HandleCtlSelectEvent(EventPtr e) FORMS_SEGMENT;
static void TrackDrag(EventPtr e) FORMS_SEGMENT;
static Boolean VT100EventHandler(struct FormGadgetTypeInCallback *gadget, UInt16 cmd, void *paramP) FORMS_SEGMENT;
static Boolean HandleCtlRepeatEvent(EventPtr e) FORMS_SEGMENT;


static Boolean OpenNetLib(void)
{
    Err err;
    UInt16 badIF = 0;
    Boolean allUp;

    if (NetLibCount == 0) {
        err = NetLibOpen(AppNetRefnum, &badIF);
        if (err  &&  err != netErrAlreadyOpen) {
            NetworkError("Could not open network library.", err);
            return false;
        } else if (badIF) {
            NetworkError("Could not open network interface.", badIF);
            return false;
        } else {
            NetLibCount++;
        }
    }

    // reconnect if necessary
    err = NetLibConnectionRefresh(AppNetRefnum, true, &allUp, &badIF);
    if (err) {
        NetworkError("Could not open network interface.", err);
        return false;
    } else if (!allUp) {
        NetworkError("Could not open network interface.", badIF);
        return false;
    } else {
        return true;
    }
}



static Boolean StartConnection(void)
{
    MemHandle recordH;
    char *hostname;
    char *portname;
    char *username;
    char *displayname;
    FieldPtr displayField;
    int port;
    int result;
    RectangleType bounds;

    ss = NULL;

    recordH=ConnectionListReadSelectedRecord(&hostname, &portname, &username);
    if (!recordH) return false;

    port = StrAToI(portname);
    if (port == 0) port = 22;

    displayField = PrvGetObjectByID(TerminalHostFieldID);
    displayname = MemPtrNew(1+StrLen(username)+StrLen("@")+StrLen(hostname));
    if (!displayname) {
        PrvSetFieldToValue(displayField, hostname);
    } else {
        StrCopy(displayname, username);
        StrCat(displayname, "@");
        StrCat(displayname, hostname);
        PrvSetFieldToValue(displayField, displayname);
        MemPtrFree(displayname);
    }

    // draw now in case of error during connection
    FrmDrawForm(FrmGetActiveForm());

    if (!OpenNetLib()) goto done;

    PrvGetObjectBoundsByID(TerminalTextGadgetID, &bounds);
    ss = ssh_create(username, bounds);
    result = ssh_open(ss, hostname, port);
    if (result < 0) {
        ssh_free(ss);
        ss = NULL;
    }

 done:
    MemHandleUnlock(recordH);
        
    return (ss != NULL);
}




static void ConstructTopVT100Pane(void)
{
    // set vt100 gadget to full size
    RectangleType bounds = {{0, 15}, {160, 120}};
    PrvSetObjectBoundsByID(TerminalTextGadgetID, &bounds);
    if (ss) ssh_set_bounds(ss, bounds);
}

static void DestructTopVT100Pane(void)
{
    // nothing to do - other pane constructors will shrink vt100 gadget
}


static void ConstructTopABCPane(void)
{
    int i;
    int start, end;
    Boolean caps, num, autoshift;
    UInt16 shift;

    // shrink vt100 gadget
    RectangleType bounds = {{0, 15}, {160, 72}};
    PrvSetObjectBoundsByID(TerminalTextGadgetID, &bounds);
    WinEraseRectangle(&bounds, 0);
    ssh_set_bounds(ss, bounds);

    // show bitmap and buttons
    PrvShowObjectByID(TerminalTABCButtonsBitmapID);
    // choose caps based on Graffiti shift state
    GrfGetState(&caps, &num, &shift, &autoshift);
    if (caps) {
        start = TerminalTABCFirstCapscaseButtonID;
        end = TerminalTABCLastCapscaseButtonID;
        PrvSetControlValue(TerminalTABCCapsButtonID, 1);
        PrvSetControlValue(TerminalTABCShiftButtonID, 0);
    } else if (shift) {
        start = TerminalTABCFirstUppercaseButtonID;
        end = TerminalTABCLastUppercaseButtonID;
        PrvSetControlValue(TerminalTABCCapsButtonID, 0);
        PrvSetControlValue(TerminalTABCShiftButtonID, 1);
    } else {
        start = TerminalTABCFirstLowercaseButtonID;
        end = TerminalTABCLastLowercaseButtonID;
        PrvSetControlValue(TerminalTABCCapsButtonID, 0);
        PrvSetControlValue(TerminalTABCShiftButtonID, 0);
    }
    for (i = start; i <= end; i++) {
        PrvShowObjectByID(i);
    }
    PrvShowObjectByID(TerminalTABCTabButtonID);
    PrvShowObjectByID(TerminalTABCSpaceButtonID);
    PrvShowObjectByID(TerminalTABCBackspaceButtonID);
    PrvShowObjectByID(TerminalTABCReturnButtonID);
    PrvShowObjectByID(TerminalTABCCapsButtonID);
    PrvShowObjectByID(TerminalTABCShiftButtonID);
}

static void DestructTopABCPane(void)
{
    int i;
    int count;

    // don't change vt100 gadget

    // hide bitmap and buttons
    PrvHideObjectByID(TerminalTABCButtonsBitmapID);
    count = TerminalTABCLastButtonID - TerminalTABCFirstButtonID + 1;
    for (i = 0; i < count; i++) {
        PrvHideObjectByID(TerminalTABCFirstButtonID + i);
    }
    PrvHideObjectByID(TerminalTABCTabButtonID);
    PrvHideObjectByID(TerminalTABCSpaceButtonID);
    PrvHideObjectByID(TerminalTABCBackspaceButtonID);
    PrvHideObjectByID(TerminalTABCReturnButtonID);
    PrvHideObjectByID(TerminalTABCCapsButtonID);
    PrvHideObjectByID(TerminalTABCShiftButtonID);
}


static void SetABCCase(case_t newcase)
{
    Boolean caps = 0, shift = 0;
    if (newcase == caseUpper) shift = 1;
    else if (newcase == caseCaps) caps = 1;

    GrfSetState(caps, 0, shift);
    GsiSetShiftState(caps ? glfCapsLock : 0, shift ? 4 : 0);
    // value 4 should be gsiShiftUpper but isn't, or something
    DestructTopABCPane();
    ConstructTopABCPane();
}


static void ConstructTop123Pane(void)
{
    int i;
    int count;

    // shrink vt100 gadget
    RectangleType bounds = {{0, 15}, {160, 72}};
    PrvSetObjectBoundsByID(TerminalTextGadgetID, &bounds);
    WinEraseRectangle(&bounds, 0);
    ssh_set_bounds(ss, bounds);

    // show bitmap and buttons
    PrvShowObjectByID(TerminalT123ButtonsBitmapID);
    count = TerminalT123LastButtonID - TerminalT123FirstButtonID + 1;
    for (i = 0; i < count; i++) {
        PrvShowObjectByID(TerminalT123FirstButtonID + i);
    }
    PrvShowObjectByID(TerminalT123TabButtonID);
    PrvShowObjectByID(TerminalT123SpaceButtonID);
    PrvShowObjectByID(TerminalT123BackspaceButtonID);
    PrvShowObjectByID(TerminalT123ReturnButtonID);
}

static void DestructTop123Pane(void)
{
    int i;
    int count;

    // don't change vt100 gadget

    // hide bitmap and buttons
    PrvHideObjectByID(TerminalT123ButtonsBitmapID);
    count = TerminalT123LastButtonID - TerminalT123FirstButtonID + 1;
    for (i = 0; i < count; i++) {
        PrvHideObjectByID(TerminalT123FirstButtonID + i);
    }
    PrvHideObjectByID(TerminalT123TabButtonID);
    PrvHideObjectByID(TerminalT123SpaceButtonID);
    PrvHideObjectByID(TerminalT123BackspaceButtonID);
    PrvHideObjectByID(TerminalT123ReturnButtonID);
}


static void ConstructTopPane(top_pane_t newPane)
{
    switch (newPane) {
    case topVT100: ConstructTopVT100Pane(); return;
    case topABC:   ConstructTopABCPane();   return;
    case top123:   ConstructTop123Pane();   return;
        // case topFkey: ConstructTopFkeyPane(); return;
        // case topEtc: ConstructTopEtcPane(); return;
    default: return;
    }
}


static void DestructTopPane(top_pane_t newPane)
{
    switch (newPane) {
    case topVT100: DestructTopVT100Pane(); return;
    case topABC:   DestructTopABCPane();   return;
    case top123:   DestructTop123Pane();   return;
        // case topFkey: DestructTopFkeyPane(); return;
        // case topEtc: DestructTopEtcPane(); return;
    default: return;
    }
}


static void SetTopPane(top_pane_t newPane)
{
    if (TopPane == newPane) {
        // nothing to do
    } else if (TopPane == topAll) {
        // Everything currently visible. Hide everything and show newPane.
        DestructTopPane(topVT100);
        DestructTopPane(topABC);
        DestructTopPane(top123);
        DestructTopPane(topFkey);
        DestructTopPane(topEtc);
        TopPane = newPane;
        ConstructTopPane(TopPane);
    } else {
        // Hide current pane and show newPane.
        DestructTopPane(TopPane);
        TopPane = newPane;
        ConstructTopPane(TopPane);
    }
}



static void ConstructBottomTCPane(void)
{
    int i;
    int count;

    PrvShowObjectByID(TerminalBTCButtonsBitmapID);
    count = TerminalBTCLastButtonID - TerminalBTCFirstButtonID + 1;
    for (i = 0; i < count; i++) {
        PrvShowObjectByID(TerminalBTCFirstButtonID + i);
    }
}

static void DestructBottomTCPane(void)
{
    int i;
    int count;

    PrvHideObjectByID(TerminalBTCButtonsBitmapID);
    count = TerminalBTCLastButtonID - TerminalBTCFirstButtonID + 1;
    for (i = 0; i < count; i++) {
        PrvHideObjectByID(TerminalBTCFirstButtonID + i);
    }
}


static void ConstructBottomFullPane(void)
{
    PrvShowObjectByID(TerminalBFullABCButtonID);
    PrvShowObjectByID(TerminalBFull123ButtonID);
    // PrvShowObjectByID(TerminalBFullFkeyButtonID);
    // PrvShowObjectByID(TerminalBFullEtcButtonID);

    PrvShowObjectByID(TerminalBFullUpButtonID);
    PrvShowObjectByID(TerminalBFullDownButtonID);
    PrvShowObjectByID(TerminalBFullLeftButtonID);
    PrvShowObjectByID(TerminalBFullRightButtonID);
}

static void DestructBottomFullPane(void)
{
    PrvHideObjectByID(TerminalBFullABCButtonID);
    PrvHideObjectByID(TerminalBFull123ButtonID);
    // PrvHideObjectByID(TerminalBFullFkeyButtonID);
    // PrvHideObjectByID(TerminalBFullEtcButtonID);

    PrvHideObjectByID(TerminalBFullUpButtonID);
    PrvHideObjectByID(TerminalBFullDownButtonID);
    PrvHideObjectByID(TerminalBFullLeftButtonID);
    PrvHideObjectByID(TerminalBFullRightButtonID);
}


static void ConstructBottomPane(bottom_pane_t newPane)
{
    switch (newPane) {
    case bottomTC:   ConstructBottomTCPane();   return;
    case bottomFull: ConstructBottomFullPane(); return;
        // case bottomCust: ConstructBottomCustPane(); return;
    default: return;
    }
}


static void DestructBottomPane(bottom_pane_t newPane)
{
    switch (newPane) {
    case bottomTC:   DestructBottomTCPane();   return;
    case bottomFull: DestructBottomFullPane(); return;
        // case bottomCust: DestructBottomCustPane(); return;
    default: return;
    }
}


static void SetBottomPane(bottom_pane_t newPane)
{
    if (BottomPane == newPane) {
        // nothing to do
    } else if (BottomPane == bottomAll) {
        // Everything currently visible. Hide everything and show newPane
        DestructBottomPane(bottomTC);
        DestructBottomPane(bottomFull);
        DestructBottomPane(bottomCust);
        BottomPane = newPane;
        ConstructBottomPane(BottomPane);
    } else {
        // Hide current pane and show newPane
        DestructBottomPane(BottomPane);
        BottomPane = newPane;
        ConstructBottomPane(BottomPane);
    }
}


static void ShowABCPanes(void)
{
    SetTopPane(topABC);
    SetBottomPane(bottomFull);
    
    PrvSetControlValue(TerminalBFullABCButtonID, 1);
    PrvSetControlValue(TerminalBFull123ButtonID, 0);
    // PrvSetControlValue(TerminalBFullFkeyButtonID, 0);
    // PrvSetControlValue(TerminalBFullEtcButtonID, 0);

    FrmDrawForm(FrmGetActiveForm());
}

static void Show123Panes(void)
{
    SetTopPane(top123);
    SetBottomPane(bottomFull);
    
    PrvSetControlValue(TerminalBFullABCButtonID, 0);
    PrvSetControlValue(TerminalBFull123ButtonID, 1);
    // PrvSetControlValue(TerminalBFullFkeyButtonID, 0);
    // PrvSetControlValue(TerminalBFullEtcButtonID, 0);

    FrmDrawForm(FrmGetActiveForm());
}


// Hide ABC, 123, Fkey, etc
// but leave bottom full intact
static void HideABC123Panes(void)
{
    SetTopPane(topVT100);
    SetBottomPane(bottomFull);
    
    PrvSetControlValue(TerminalBFullABCButtonID, 0);
    PrvSetControlValue(TerminalBFull123ButtonID, 0);
    // PrvSetControlValue(TerminalBFullFkeyButtonID, 0);
    // PrvSetControlValue(TerminalBFullEtcButtonID, 0);

    FrmDrawForm(FrmGetActiveForm());
}




static char control_character(char c)
{
    if (c == ' '  ||  c == '@') return '\0';        // C-space, C-@
    if (c >= 'a'  &&  c <= 'z') return c - 'a' + 1; // C-a .. C-z
    if (c >= 'A'  &&  c <= 'Z') return c - 'A' + 1; // C-A .. C-Z
    if (c >= '['  &&  c <= '_') return c - 'A' + 1; // C-[ .. C-_
    if (c >= '{'  &&  c <= '}') return c - 'a' + 1; // C-{ .. C-} (no C-~)
    return c;
}


static void SendModifiedKey(struct ssh_session_t *ss, unsigned char c, 
                            int ctrlPressed, int metaPressed)
{
    if (ctrlPressed) {
        c = control_character(c);
    }
    if (metaPressed) {
        char bytes[2];
        bytes[0] = '\e'; // ESC prefix
        bytes[1] = c;
        ssh_keys(ss, bytes, 2);
    } else {
        ssh_key(ss, c);
    }
}


static void ClearModifiers(void)
{
    PrvSetControlValue(TerminalCtrlButtonID, 0);
    PrvSetControlValue(TerminalMetaButtonID, 0);
    if (TopPane == topABC  &&  PrvGetControlValue(TerminalTABCShiftButtonID)) {
        SetABCCase(caseLower);
    }
}

static void SendModifiedKeyAndClearModifiers(unsigned char c)
{
    SendModifiedKey(ss, c, 
                    PrvGetControlValue(TerminalCtrlButtonID), 
                    PrvGetControlValue(TerminalMetaButtonID));
    ClearModifiers();
}


static void SendUpArrowKey(void)
{
    ssh_keys(ss, "\eOA", 3);
}

static void SendDownArrowKey(void)
{
    ssh_keys(ss, "\eOB", 3);
}

static void SendRightArrowKey(void)
{
    ssh_keys(ss, "\eOC", 3);
}

static void SendLeftArrowKey(void)
{
    ssh_keys(ss, "\eOD", 3);
}


static Boolean HandleCtlSelectEvent(EventPtr e)
{
    UInt16 id = e->data.ctlSelect.controlID;

    // Check for self-labeled controls
    if ((id >= TerminalBTCFirstButtonID  && id <= TerminalBTCLastButtonID)  ||
        (id >= TerminalTABCFirstButtonID && id <= TerminalTABCLastButtonID) ||
        (id >= TerminalT123FirstButtonID && id <= TerminalT123LastButtonID))
    {
        const char *text = CtlGetLabel(e->data.ctlSelect.pControl);
        if (text && text[0]) SendModifiedKeyAndClearModifiers(text[0]);
        return false;
    }

    switch (id) {
    case TerminalCtrlButtonID:
    case TerminalMetaButtonID:
        // nothing to do
        return false;

    case TerminalTABCTabButtonID:
    case TerminalT123TabButtonID:
        SendModifiedKeyAndClearModifiers('\t');
        return false;

    case TerminalTABCSpaceButtonID:
    case TerminalT123SpaceButtonID:
        SendModifiedKeyAndClearModifiers(' ');
        return false;

    case TerminalTABCBackspaceButtonID:
    case TerminalT123BackspaceButtonID:
        SendModifiedKeyAndClearModifiers(127); // DEL not BS
        return false;

    case TerminalTABCReturnButtonID:
    case TerminalT123ReturnButtonID:
        SendModifiedKeyAndClearModifiers(13); // CR not LF
        return false;

    case TerminalTABCShiftButtonID:
        if (PrvGetControlValue(TerminalTABCShiftButtonID)) {
            SetABCCase(caseUpper);
        } else {
            SetABCCase(caseLower);
        }
        return false;

    case TerminalTABCCapsButtonID:
        if (PrvGetControlValue(TerminalTABCCapsButtonID)) {
            SetABCCase(caseCaps);
        } else {
            SetABCCase(caseLower);
        }
        return false;

    case TerminalBFullABCButtonID:
        if (PrvGetControlValue(TerminalBFullABCButtonID)) {
            ShowABCPanes();
        } else {
            HideABC123Panes();
        }
        return false;

    case TerminalBFull123ButtonID:
        if (PrvGetControlValue(TerminalBFull123ButtonID)) {
            Show123Panes();
        } else {
            HideABC123Panes();
        }
        return false;

    default:
        return false;
    }
}


static void TrackDrag(EventPtr e)
{
    Boolean moved = 0;
    Boolean stillDown;
    Coord x, y;
    static RectangleType slopRect = {{0, 0}, {0, 0}};
    static int clickCount = 0;
    static UInt32 lastClickTime = 0;
    UInt32 now;

    // fixme drag stops all network events

    x = e->screenX;
    y = e->screenY;

    // See whether this click counts as a multi-click
    now = TimGetTicks();
    if (RctPtInRectangle(x, y, &slopRect)  &&  
        now - lastClickTime < SysTicksPerSecond() / 3)
    {
        clickCount++;
        // keep same slop rect
    } else {
        clickCount = 1;
        slopRect.topLeft.x = x - 1;
        slopRect.topLeft.y = y - 1;
        slopRect.extent.x = 3;
        slopRect.extent.y = 3;
    }
    lastClickTime = now;

    // click
    ssh_click(ss, x, y, clickCount);
    ssh_click(ss, x, y, -1);

    while (true) {
        SysTaskDelay(SysTicksPerSecond() / 30); // don't spin too fast
        PenGetPoint(&x, &y, &stillDown);
        if (!stillDown) break;

        if (!moved) moved = !RctPtInRectangle(x, y, &slopRect);
        if (moved) {
            // drag
            ssh_click(ss, x, y, -1);
        }
    }

    // release
    ssh_click(ss, x, y, 0);

}

static Boolean VT100EventHandler(struct FormGadgetTypeInCallback *gadget, 
                                 UInt16 cmd, void *paramP)
{
    Boolean handled = false; 

    switch (cmd) { 
    case formGadgetDrawCmd: 
        // Sent to active gadgets any time form is  
        // drawn or redrawn. 
        if (ss) ssh_activate(ss);
        handled = true; 
        break; 

    case formGadgetHandleEventCmd: {
        EventPtr e = (EventPtr)paramP;
        // Sent when form receives a gadget event.  
        // paramP points to EventType structure.  
        if (e->eType == frmGadgetEnterEvent) { 
            // penDown in gadget's bounds.  
            UInt16 index;
            FormPtr frmP = FrmGetActiveForm();
            if ((index = FrmGetFocus(frmP)) != noFocus) {
                // selection in vt100 disables selection in real fields
                // fixme this might fail if there were a focused table
                FldSetSelection((FieldPtr)FrmGetObjectPtr(frmP, index), 0, 0);
                FrmSetFocus(frmP, noFocus);
            }
            TrackDrag(e);
            handled = true; 
        } else if (e->eType == frmGadgetMiscEvent) { 
            // This event is sent by your application 
            // when it needs to send info to the gadget 
        } 
        break; 
    }

    case formGadgetDeleteCmd:  
        // Perform any cleanup prior to deletion. 
        break; 

    case formGadgetEraseCmd:  
        // FrmHideObject takes care of this if you  
        // return false.  
        handled = false; 
        break; 

    default:
        handled = false;
        break;
    } 

    return handled; 
} 


static Boolean HandleCtlRepeatEvent(EventPtr e)
{
    switch (e->data.ctlSelect.controlID) {
    case TerminalRScrollPageUpButtonID:
        if (ss) ssh_scroll(ss, -(ssh_visible_height(ss)-1));
        return false;

    case TerminalRScrollPageDownButtonID:
        if (ss) ssh_scroll(ss, ssh_visible_height(ss)-1);
        return false;
        
    case TerminalRScrollLineUpButtonID:
        if (ss) ssh_scroll(ss, -1);
        return false;
        
    case TerminalRScrollLineDownButtonID:
        if (ss) ssh_scroll(ss, 1);
        return false;

        // fixme these repeating buttons don't work as smoothly as the 5-way
    case TerminalBFullUpButtonID:
        // case TerminalRArrowsUpButtonID:
        ClearModifiers();
        SendUpArrowKey();
        return false;

    case TerminalBFullDownButtonID:
        // case TerminalRArrowsDownButtonID:
        ClearModifiers();
        SendDownArrowKey();
        return false;

    case TerminalBFullLeftButtonID:
        // case TerminalRArrowsLeftButtonID:
        ClearModifiers();
        SendLeftArrowKey();
        return false;

    case TerminalBFullRightButtonID:
        // case TerminalRArrowsRightButtonID:
        ClearModifiers();
        SendRightArrowKey();
        return false;

    default: 
        return false;
    }
}

Boolean TerminalFormHandleEvent(EventPtr e)
{
    switch (e->eType) {
    case frmOpenEvent:
        FrmSetGadgetHandler(FrmGetActiveForm(), 
                            PrvGetObjectIndexByID(TerminalTextGadgetID), 
                            VT100EventHandler);
        // fixme pane preferences
        SetTopPane(topVT100);
        SetBottomPane(bottomTC);
        //SetRightPane(rightScroll);
        FrmDrawForm(FrmGetActiveForm());

        if (!StartConnection()) {
            FrmGotoForm(MainFormID);
            return true;
        }

        FrmDrawForm(FrmGetActiveForm());
        return true;

    case menuEvent:
        switch (e->data.menu.itemID) {
        case MenuDisconnect:
            FrmGotoForm(MainFormID);
            return true;
        case MenuAboutPalmSSH:
            About();
            return true;
        case MenuCredits:
            FrmPopupForm(CreditsFormID);
            return true;
        case sysEditMenuUndoCmd:
            // no undo in vt100 - send to fields
            return false;
        case sysEditMenuCutCmd:
            // no cut in vt100 - copy will do
            // FALL-THROUGH to copy
        case sysEditMenuCopyCmd:
            if (FrmGetFocus(FrmGetActiveForm()) == noFocus) {
                ssh_copy(ss);
                // don't send to other fields
                return true; 
            } else {
                return false;
            }
        case sysEditMenuPasteCmd:
            if (FrmGetFocus(FrmGetActiveForm()) == noFocus) {
                ssh_paste(ss);
                // don't send to other fields
                return true;
            } else {
                // let real fields handle it
                return false;
            }
        case sysEditMenuKeyboardCmd:
            ShowABCPanes();
            return true;
        case sysEditMenuGraffitiCmd:
            SysGraffitiReferenceDialog(referenceDefault);
            return true;

        default:
            return false;
        }

    case menuCmdBarOpenEvent: {
        UInt16 len;
        // fixme buttons get added in wrong order if hostname field has 
        // a selection (need to override default field buttons)
        // add "paste" button if there is clipboard content
        if (ClipboardGetItem(clipboardText, &len)) {
            MenuCmdBarAddButton(menuCmdBarOnLeft, BarPasteBitmap, menuCmdBarResultMenuItem, sysEditMenuPasteCmd, NULL);
        }
        // add "copy" button if there is a terminal selection
        if (ssh_selection_exists(ss)) {
            MenuCmdBarAddButton(menuCmdBarOnLeft, BarCopyBitmap, menuCmdBarResultMenuItem, sysEditMenuCopyCmd, NULL);
        }
        return false;
    }

    case frmCloseEvent:
        if (ss) {
            ssh_close(ss);
            ssh_free(ss);
            ss = NULL;
        }
        TopPane = -1;
        BottomPane = -1;
        RightPane = -1;
        return false;

    case fldEnterEvent:
        // selection in real field disables selection in vt100
        if (ss) {
            ssh_deselect(ss);
        }
        return false;

    case ctlRepeatEvent:
        return HandleCtlRepeatEvent(e);
        
    case ctlSelectEvent:
        return HandleCtlSelectEvent(e);

    case keyDownEvent: {
        if (!EvtKeydownIsVirtual(e)) {
            unsigned char c = e->data.keyDown.chr;
            
            // send CR rather than LF
            if (c == 10) c = 13;
            // send DEL rather than BS
            if (c == 8) c = 127;

            SendModifiedKeyAndClearModifiers(c);

            return true;
        } 
        else if (e->data.keyDown.chr == vchrKeyboard)
        {
            ShowABCPanes();
        }
        else if (e->data.keyDown.chr == vchrKeyboardAlpha)
        {
            ShowABCPanes();
        }
        else if (e->data.keyDown.chr == vchrKeyboardNumeric)
        {
            Show123Panes();
        }
        else if (e->data.keyDown.chr == 0x51d)
        {
            // Tungsten|C fn-space (usually runs sys keyboard in symbol pane)
            // Here toggle between full(123) and T|C
            if (BottomPane != bottomFull) {
                Show123Panes();
            } else {
                SetTopPane(topVT100);
                SetBottomPane(bottomTC);
                FrmDrawForm(FrmGetActiveForm());
            }
        }
        else if (NavSelectPressed(e)) {
            // nav button == toggle ctrl->meta->both->neither
            int ctrlWasPressed = PrvGetControlValue(TerminalCtrlButtonID);
            int metaWasPressed = PrvGetControlValue(TerminalMetaButtonID);

            if (ctrlWasPressed) {
                if (metaWasPressed) {
                    PrvSetControlValue(TerminalCtrlButtonID, 0);
                    PrvSetControlValue(TerminalMetaButtonID, 0);
                } else {
                    PrvSetControlValue(TerminalCtrlButtonID, 0);
                    PrvSetControlValue(TerminalMetaButtonID, 1);
                }
            } else {
                if (metaWasPressed) {
                    PrvSetControlValue(TerminalCtrlButtonID, 1);
                    PrvSetControlValue(TerminalMetaButtonID, 1);
                } else {
                    PrvSetControlValue(TerminalCtrlButtonID, 1);
                    PrvSetControlValue(TerminalMetaButtonID, 0);
                }
            }
            return true;
        }
        // fixme ctrl-arrow = page?
        else if (NavDirectionPressed(e, Up)) {
            ClearModifiers();
            SendUpArrowKey();
            return true;
        } else if (NavDirectionPressed(e, Down)) {
            ClearModifiers();
            SendDownArrowKey();
            return true;
        } else if (NavDirectionPressed(e, Right)) {
            ClearModifiers();
            SendRightArrowKey();
            return true;
        } else if (NavDirectionPressed(e, Left)) {
            ClearModifiers();
            SendLeftArrowKey();
            return true;
        }
        return false;
    }

    case winExitEvent:
        if ((FormPtr)e->data.winEnter.enterWindow == 
            FrmGetFormPtr(TerminalFormID))
        {
            if (ss) ssh_deactivate(ss);
            return true;
        }
        return false;

    case winEnterEvent:
        if ((FormPtr)e->data.winEnter.enterWindow == 
            FrmGetFormPtr(TerminalFormID))
        {
            if (ss) ssh_activate(ss);
            return true;
        }
        return false;
        
    default:
        return false;
    }
}


// Terminal form wants to catch keyboard vchrs to display its own keyboard
// fixme need to test "abc" and "123" on a real device (they weren't working 
// at one point, but then this code was changed)
Boolean TerminalFormStealEvent(EventPtr e)
{
    if (e->eType != keyDownEvent) return false;
    if ((e->data.keyDown.modifiers & commandKeyMask) == 0) return false;
    if (e->data.keyDown.chr != vchrKeyboard  &&  
        e->data.keyDown.chr != vchrKeyboardAlpha  &&  
        e->data.keyDown.chr != vchrKeyboardNumeric  &&  
        e->data.keyDown.chr != 0x51d) return false;
    // it's a keyboard-display event
    if (FrmGetActiveFormID() != TerminalFormID) return false;
    // Terminal form is open
    // if (FrmGetFocus(FrmGetActiveForm()) != noFocus) return false;
    // No focused field
    // fixme wrong because we do want keyboard if hostname field is selected

    // steal it
    return true;
}
