/**********
 * Copyright (c) 2003-2004 Greg Parker.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY GREG PARKER ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **********/

#include "includes.h"
#include "connection.h"

#include "session.h"
#include "transport.h"
#include "packetizer.h"
#include "vt100/vt100.h"
#include "ssh2.h"
#include "openssh/buffer.h"
#include "openssh/bufaux.h"
#include "openssh/match.h"
#include "xmalloc.h"

#define DATA_MAX 32768
#define RECEIVE_WINDOW_MIN DATA_MAX
#define RECEIVE_WINDOW_MAX DATA_MAX*4


static void connection_send_pty_request(ssh_session_t *ss, uint32_t channel) CONNECTION_SEGMENT;
static void connection_send_shell_request(ssh_session_t *ss, uint32_t channel) CONNECTION_SEGMENT;
static void connection_send_service_request(ssh_session_t *ss, char *name) CONNECTION_SEGMENT;
static void connection_send_userauth(ssh_session_t *ss, char *username, char *authmethod, uint8_t *authdata, uint16_t authdatalen) CONNECTION_SEGMENT;
static void connection_send_userauth_password(ssh_session_t *ss, char *username, char *password) CONNECTION_SEGMENT;
static void connection_send_authmethod_request(ssh_session_t *ss, char *username) CONNECTION_SEGMENT;
static void connection_send_channel_open(ssh_session_t *ss, uint32_t channel, uint32_t receiveWindow, uint32_t packetMax) CONNECTION_SEGMENT;
static void connection_send_window_adjust(ssh_session_t *ss, uint32_t channel, uint32_t adjustment) CONNECTION_SEGMENT;
static void connection_send_channel_close(ssh_session_t *ss, uint32_t channel) CONNECTION_SEGMENT;
static void connection_send_global_request_failure(ssh_session_t *ss) CONNECTION_SEGMENT;
static void connection_send_channel_request_failure(ssh_session_t *ss, uint32_t channel) CONNECTION_SEGMENT;
static void connection_send_channel_data(ssh_session_t *ss, uint32_t channel, uint8_t *bytes, uint16_t len) CONNECTION_SEGMENT;
static connection_state_t connection_get_password(ssh_session_t *ss) CONNECTION_SEGMENT;
static void connection_increase_receive_window(ssh_session_t *ss) CONNECTION_SEGMENT;
static void connection_handle_typed_data(ssh_session_t *ss, uint32_t type, uint8_t *data, uint32_t datalen, int closing) CONNECTION_SEGMENT;
static connection_state_t connection_handle_unexpected_message(ssh_session_t *ss, uint8_t msg, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_handle_userauth_service_accept(ssh_session_t *ss, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_handle_banner(ssh_session_t *ss, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_handle_userauth_success(ssh_session_t *ss, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_handle_userauth_failure(ssh_session_t *ss, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_handle_password_change(ssh_session_t *ss, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_handle_channel_open(ssh_session_t *ss, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_handle_channel_open_failure(ssh_session_t *ss, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_handle_pty_success(ssh_session_t *ss, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_handle_pty_failure(ssh_session_t *ss, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_handle_shell_success(ssh_session_t *ss, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_handle_shell_failure(ssh_session_t *ss, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_handle_window_adjust(ssh_session_t *ss, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_handle_channel_close(ssh_session_t *ss, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_handle_data(ssh_session_t *ss, Buffer *packet, int closing) CONNECTION_SEGMENT;
static connection_state_t connection_handle_extended_data(ssh_session_t *ss, Buffer *packet, int closing) CONNECTION_SEGMENT;
static connection_state_t connection_state_waiting_for_userauth_service(ssh_session_t *ss, uint8_t msg, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_state_waiting_for_password(ssh_session_t *ss, uint8_t msg, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_state_waiting_for_userauth(ssh_session_t *ss, uint8_t msg, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_state_waiting_for_channel_open(ssh_session_t *ss, uint8_t msg, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_state_waiting_for_pty(ssh_session_t *ss, uint8_t msg, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_state_waiting_for_shell(ssh_session_t *ss, uint8_t msg, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_state_running(ssh_session_t *ss, uint8_t msg, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_state_closing(ssh_session_t *ss, uint8_t msg, Buffer *packet) CONNECTION_SEGMENT;
static connection_state_t connection_state_closed(ssh_session_t *ss, uint8_t msg, Buffer *packet) CONNECTION_SEGMENT;
static void connection_cleanup(ssh_session_t *ss) CONNECTION_SEGMENT;


/***********************************************************************
 * raw packet output
 */

static void connection_send_pty_request(ssh_session_t *ss, uint32_t channel)
{
    Buffer payload;

    printf("   Opening pty...\r\n");

    buffer_init(&payload);
    buffer_put_char(&payload, SSH2_MSG_CHANNEL_REQUEST);
    buffer_put_int(&payload, channel);
    buffer_put_cstring(&payload, "pty-req");
    buffer_put_char(&payload, 1); // want-reply
    buffer_put_cstring(&payload, "xterm"); // TERM (fixme configurable)
    // fixme hardcoded sizes
    buffer_put_int(&payload, 80); // tty width  (chars)
    buffer_put_int(&payload, 40); // tty height (chars)
    buffer_put_int(&payload, 0);  // tty width  (pixels)
    buffer_put_int(&payload, 0);  // tty height (pixels)
    buffer_put_cstring(&payload, ""); // encoded terminal modes
    transport_send_payload(ss, &payload);
    buffer_free(&payload);
    connection_log("connection: sent pty request\r\n");
}


static void connection_send_shell_request(ssh_session_t *ss, uint32_t channel)
{
    Buffer payload;

    printf("   Starting shell...\r\n");

    buffer_init(&payload);
    buffer_put_char(&payload, SSH2_MSG_CHANNEL_REQUEST);
    buffer_put_int(&payload, channel);
    buffer_put_cstring(&payload, "shell");
    buffer_put_char(&payload, 1); // want-reply
    transport_send_payload(ss, &payload);
    buffer_free(&payload);
    connection_log("connection: sent shell request\r\n");
}


static void connection_send_window_change(ssh_session_t *ss, uint32_t channel, 
                                          int charsWide, int charsHigh, 
                                          int pixelsWide, int pixelsHigh)
{
    Buffer payload;

    buffer_init(&payload);
    buffer_put_char(&payload, SSH2_MSG_CHANNEL_REQUEST);
    buffer_put_int(&payload, channel);
    buffer_put_cstring(&payload, "window-change");
    buffer_put_char(&payload, 0); // want-reply
    buffer_put_int(&payload, charsWide);
    buffer_put_int(&payload, charsHigh);
    buffer_put_int(&payload, pixelsWide);
    buffer_put_int(&payload, pixelsHigh);
    transport_send_payload(ss, &payload);
    buffer_free(&payload);
    connection_log("connection: sent WINCH\r\n");    
}

static void connection_send_service_request(ssh_session_t *ss, char *name)
{
    Buffer packet;
    buffer_init(&packet);
    buffer_put_char(&packet, SSH2_MSG_SERVICE_REQUEST);
    buffer_put_cstring(&packet, name);
    transport_send_payload(ss, &packet);
    buffer_free(&packet);
    connection_log("connection: sent service request for '%s'\r\n", name);
}


static void 
connection_send_userauth(ssh_session_t *ss, char *username, 
                         char *authmethod, uint8_t *authdata, 
                         uint16_t authdatalen)
{
    Buffer payload;

    buffer_init(&payload);
    buffer_put_char(&payload, SSH2_MSG_USERAUTH_REQUEST);
    buffer_put_cstring(&payload, username);
    buffer_put_cstring(&payload, "ssh-connection"); // service
    buffer_put_cstring(&payload, authmethod);
    if (authdata) {
        buffer_append(&payload, authdata, authdatalen);
    }
    transport_send_payload(ss, &payload);
    buffer_free(&payload);
    connection_log("connection: sent userauth request for username '%s' using authmethod '%s'\r\n", username, authmethod);
}


static void 
connection_send_userauth_password(ssh_session_t *ss, char *username, char *password)
{
    Buffer authdata;

    buffer_init(&authdata);
    buffer_put_char(&authdata, 0); // boolean FALSE
    buffer_put_cstring(&authdata, password);
    connection_send_userauth(ss, username, "password", 
                             buffer_ptr(&authdata), buffer_len(&authdata));
    buffer_free(&authdata);
}


static void 
connection_send_authmethod_request(ssh_session_t *ss, char *username)
{
    // To request the list of authentication methods, 
    // send a USERAUTH_REQUEST with authentication type "none". 
    // This will either succeed with no auth (USERAUTH_SUCCESS), 
    // or be rejected with a list of authentication methods supported.

    connection_send_userauth(ss, username, "none", NULL, 0);
}


static void 
connection_send_channel_open(ssh_session_t *ss, uint32_t channel, 
                             uint32_t receiveWindow, uint32_t packetMax)
{
    Buffer payload;

    printf("   Opening channel...\r\n");

    buffer_init(&payload);
    buffer_put_char(&payload, SSH2_MSG_CHANNEL_OPEN);
    buffer_put_cstring(&payload, "session");
    buffer_put_int(&payload, channel);
    buffer_put_int(&payload, receiveWindow); // initial window size
    buffer_put_int(&payload, packetMax); // packet size
    transport_send_payload(ss, &payload);
    buffer_free(&payload);
    connection_log("connection: sent channel open\r\n");
}


static void 
connection_send_window_adjust(ssh_session_t *ss, uint32_t channel, 
                              uint32_t adjustment)
{
    // this gets used a lot, so keep a static Buffer
    static Buffer payload;
    if (!payload.buf) buffer_init(&payload);
    buffer_clear(&payload);
    buffer_put_char(&payload, SSH2_MSG_CHANNEL_WINDOW_ADJUST);
    buffer_put_int(&payload, channel);
    buffer_put_int(&payload, adjustment);
    transport_send_payload(ss, &payload);
    connection_log("connection: sent window adjust\r\n");
}


static void 
connection_send_channel_close(ssh_session_t *ss, uint32_t channel)
{
    Buffer payload;
    buffer_init(&payload);
    buffer_put_char(&payload, SSH2_MSG_CHANNEL_CLOSE);
    buffer_put_int(&payload, channel);
    transport_send_payload(ss, &payload);
    buffer_free(&payload);
    connection_log("connection: sent channel close\r\n");
}

static void 
connection_send_global_request_failure(ssh_session_t *ss)
{
    Buffer payload;
    buffer_init(&payload);
    buffer_put_char(&payload, SSH2_MSG_REQUEST_FAILURE);
    transport_send_payload(ss, &payload);
    buffer_free(&payload);
    connection_log("connection: sent global request failure\r\n");
}


static void 
connection_send_channel_request_failure(ssh_session_t *ss, uint32_t channel)
{
    Buffer payload;
    buffer_init(&payload);
    buffer_put_char(&payload, SSH2_MSG_CHANNEL_FAILURE);
    buffer_put_int(&payload, channel);
    transport_send_payload(ss, &payload);
    buffer_free(&payload);
    connection_log("connection: sent channel request failure\r\n");
}


// does NOT enforce send window and max packet size
static void 
connection_send_channel_data(ssh_session_t *ss, uint32_t channel, 
                             uint8_t *bytes, uint16_t len)
{
    static Buffer payload;
    if (!payload.buf) buffer_init(&payload);
    buffer_clear(&payload);
    buffer_put_char(&payload, SSH2_MSG_CHANNEL_DATA);
    buffer_put_int(&payload, channel);
    buffer_put_string(&payload, bytes, len);
    transport_send_payload(ss, &payload);
}


/***********************************************************************
 * misc
 */

static connection_state_t
connection_get_password(ssh_session_t *ss)
{
    ss->c.passwordAttempts++;

    if (ss->c.passwordAttempts == 1) {
        // Look for a saved username and password, but only once
        // fixme wait for banner?
        // fixme
    }

    if (ss->c.passwordAttempts > 3) {
        // too many password attempts
        // fixme don't bother enforcing this?
    }

    // get name and password from GUI
    // reuse username if known
    // fixme
    ssh_request_password(ss->c.passwordAttempts);

    return CONNECTION_STATE_WAITING_FOR_PASSWORD;
}


void connection_use_password(ssh_session_t *ss, char *username, char *password)
{
    if (ss->c.state == CONNECTION_STATE_WAITING_FOR_PASSWORD) {
        if (ss->username) arena_free(ss->username);
        ss->username = arena_strdup(username);
        connection_send_userauth_password(ss, username, password);
        ss->c.state = CONNECTION_STATE_WAITING_FOR_USERAUTH;
    } else {
        printf("connection: got password outside waiting-for-password\r\n");
    }
}

void connection_use_bounds(ssh_session_t *ss, RectangleType bounds, 
                           int charsWide, int charsHigh)
{
    if (ss->c.state == CONNECTION_STATE_RUNNING) {
        connection_send_window_change(ss, ss->c.remoteChannel, 
                                      charsWide, charsHigh, 
                                      bounds.extent.x, bounds.extent.y);
    } else {
        // fixme need to send it later
    }
}

static void connection_increase_receive_window(ssh_session_t *ss)
{
    uint32_t delta;
    if (ss->c.receiveWindow < RECEIVE_WINDOW_MIN) {
        delta = RECEIVE_WINDOW_MAX - ss->c.receiveWindow;
        connection_send_window_adjust(ss, ss->c.remoteChannel, delta);
        ss->c.receiveWindow += delta;
    }
}


static void connection_handle_typed_data(ssh_session_t *ss, uint32_t type, 
                                         uint8_t *data, uint32_t datalen, 
                                         int closing)
{
    if (ss->c.receiveWindow < datalen) {
        printf("connection: receive window full\r\n");
        return; // ignore data
    }
    ss->c.receiveWindow -= datalen;
    if (ss->c.receiveWindow < RECEIVE_WINDOW_MIN  &&  !closing) {
        connection_increase_receive_window(ss);
    }

    if (type == 0  ||  type == 1) {
        // fixme handle stderr differently?
        ssh_receive_bytes(ss, data, datalen);
    } else {
        printf("connection: got %d bytes of data typed %d\r\n", datalen, type);
    }
}



/***********************************************************************
 * incoming message handlers
 */

static connection_state_t 
connection_handle_unexpected_message(ssh_session_t *ss, 
                                     uint8_t msg, Buffer *packet)
{
    if (msg == SSH2_MSG_GLOBAL_REQUEST) {
        // all global requests are refused
        uint16_t len;
        uint8_t *request;
        uint8_t want_reply;
        request = buffer_get_string(packet, &len);
        want_reply = buffer_get_char(packet);
        // do NOT require empty - there might be unknown request-specific data
        printf("connection (state %d): refused global request '%s'\r\n", 
                   ss->c.state, request);
        arena_free(request);
        if (want_reply) {
            connection_send_global_request_failure(ss);
        }
        return ss->c.state;
    } else if (msg == SSH2_MSG_CHANNEL_REQUEST) {
        // all channel requests are currently refused with a printf
        // EXCEPT: keepalive@openssh.com and exit-status are silently refused
        // fixme requests that may be desirable (would implement elsewhere):
        // xon-xoff, signal, exit-status, exit-signal, keepalive@openssh.com
        uint16_t len;
        uint32_t channel;
        uint8_t *request;
        uint8_t want_reply;
        channel = buffer_get_int(packet);
        request = buffer_get_string(packet, &len);
        want_reply = buffer_get_char(packet);
        // do NOT require empty - there might be unknown request-specific data
        // Don't print keepalives or exit-status
        if (0 != StrCompare(request, "keepalive@openssh.com")  &&  
            0 != StrCompare(request, "exit-status"))
        {
            printf("connection (state %d): refused channel request '%s'\r\n",
                   ss->c.state, request);
        }
        arena_free(request);
        if (ss->c.localChannel == 0  ||  channel != ss->c.localChannel) {
            // wrong channel or no channel - die
            ssh_kill(ss);
            return ss->c.state;
        }
        if (want_reply) {
            connection_send_channel_request_failure(ss, ss->c.remoteChannel);
        }
        return ss->c.state;
    } else {
        // fixme die?
        // fixme are there any other messages?
        // fixme send UNIMPLEMENTED?
        printf("connection (state %d): ignored unknown message %d\r\n", 
                   ss->c.state, msg);
        return ss->c.state;
    }
}


static connection_state_t 
connection_handle_userauth_service_accept(ssh_session_t *ss, Buffer *packet)
{
    uint16_t len;
    char *service = buffer_get_string(packet, &len);
    if (len != strlen("ssh-userauth")  ||  
        0 != memcmp(service, "ssh-userauth", len)) 
    {
        printf("connection: service-accept for wrong service '%s'\r\n",
                   service);
        arena_free(service);
        ssh_kill(ss);
        return ss->c.state;
    }

    arena_free(service);
    buffer_require_empty(packet);

    printf("   Authenticating (none) ... ");

    connection_send_authmethod_request(ss, ss->username);
    return CONNECTION_STATE_WAITING_FOR_USERAUTH;
}


static connection_state_t 
connection_handle_banner(ssh_session_t *ss, Buffer *packet)
{
    uint16_t len;
    char *msg = buffer_get_string(packet, &len);
    buffer_require_empty(packet);
    printf("banner '%s'\r\n", msg);
    arena_free(msg);
    return ss->c.state;
}


static connection_state_t 
connection_handle_userauth_success(ssh_session_t *ss, Buffer *packet)
{
    buffer_require_empty(packet);

    printf("succeeded\r\n"); // terminate "Authenticating..." line

    ss->c.localChannel = 42;
    ss->c.receiveWindow = RECEIVE_WINDOW_MAX;
    
    connection_send_channel_open(ss, ss->c.localChannel, 
                                 ss->c.receiveWindow, DATA_MAX);

    return CONNECTION_STATE_WAITING_FOR_CHANNEL_OPEN;    
}


static connection_state_t 
connection_handle_userauth_failure(ssh_session_t *ss, Buffer *packet)
{
    uint16_t len;
    uint8_t partialSuccess;
    uint8_t *usableAuthentications = NULL;
    char *authmethod_name = NULL;
    connection_state_t result;

    // terminate previous "Authenticating ..." line
    printf("failed\r\n");

    usableAuthentications = buffer_get_string(packet, &len);
    partialSuccess = buffer_get_char(packet);
    buffer_require_empty(packet);

    if (partialSuccess) {
        // "more authentications are needed" - not supported, bail
        printf("userauth: server requires more than simple password authentication (server offered '%s')\r\n", usableAuthentications);
        ssh_kill(ss);
        result = ss->c.state;
        goto done;
    }

    // Try password authentication (perhaps again)
    // fixme support other auth methods here
    authmethod_name = match_list("password", usableAuthentications, NULL);
    if (! authmethod_name) {
        // server does not support password auth - bail
        // fixme do something else if password has been tried once
        printf("userauth: server does not support password authentication (server offered '%s'\r\n", usableAuthentications);
        ssh_kill(ss);
        result = ss->c.state;
        goto done;
    } 
    else {
        // try password auth
        printf("   Authenticating (password) ... ");

        result = connection_get_password(ss);
        goto done;
    }

 done:
    if (usableAuthentications) arena_free(usableAuthentications);
    if (authmethod_name) arena_free(authmethod_name);
    return result;
}


static connection_state_t 
connection_handle_password_change(ssh_session_t *ss, Buffer *packet)
{
    // fixme
    /*
    uint16_t promptLen, languageLen;
    uint8_t *prompt = buffer_get_string(packet, &promptLen);
    uint8_t *language = buffer_get_string(packet, &languageLen);
    arena_free(language);
    buffer_require_empty(packet);
    connection_get_new_password(ss, prompt, promptLen);
    arena_free(prompt);
    return CONNECTION_STATE_WAITING_FOR_NEW_PASSWORD;
    */
    printf("connection: password change unimplemented\r\n");
    return ss->c.state;
}


static connection_state_t 
connection_handle_channel_open(ssh_session_t *ss, Buffer *packet)
{
    uint32_t recipientChannel;
    uint32_t senderChannel;
    uint32_t windowSize;
    uint32_t packetSize;
    
    recipientChannel = buffer_get_int(packet);
    if (recipientChannel != ss->c.localChannel) {
        printf("connection: channel-open-confirmation had unexpected recipient channel (wanted %d, got %d)\r\n", ss->c.localChannel, recipientChannel);
        ssh_kill(ss);
        return ss->c.state;
    }

    senderChannel = buffer_get_int(packet);
    windowSize = buffer_get_int(packet);
    packetSize = buffer_get_int(packet);
    buffer_require_empty(packet);

    ss->c.remoteChannel = senderChannel;
    ss->c.sendWindow = windowSize;
    ss->c.sendMaxPacket = packetSize;
    ss->c.channelOpen = 1;

    connection_send_pty_request(ss, ss->c.remoteChannel);
    return CONNECTION_STATE_WAITING_FOR_PTY;
}


static connection_state_t 
connection_handle_channel_open_failure(ssh_session_t *ss, Buffer *packet)
{
    uint32_t channel;
    uint32_t reasonCode;
    uint16_t len;
    uint8_t *reasonString;
    uint16_t languageLen;
    uint8_t *languageString;

    channel = buffer_get_int(packet);
    // don't bother verifying channel - we're about to die anyway

    reasonCode = buffer_get_int(packet);
    reasonString = buffer_get_string(packet, &len);
    languageString = buffer_get_string(packet, &languageLen);
    // don't bother requiring empty packet - we're about to die anyway

    printf("connection: channel-open-failure because %d '%s'\r\n", 
               reasonCode, reasonString);
    arena_free(reasonString);
    arena_free(languageString);

    ssh_kill(ss);
    return ss->c.state;
}


static connection_state_t 
connection_handle_pty_success(ssh_session_t *ss, Buffer *packet)
{
    uint32_t channel = buffer_get_int(packet);
    if (channel != ss->c.localChannel) {
        printf("connection: pty success on wrong channel (wanted %d, got %d)\r\n", ss->c.localChannel, channel);
        ssh_kill(ss);
        return ss->c.state;
    }

    buffer_require_empty(packet);

    connection_send_shell_request(ss, ss->c.remoteChannel);
    return CONNECTION_STATE_WAITING_FOR_SHELL;
}


static connection_state_t 
connection_handle_pty_failure(ssh_session_t *ss, Buffer *packet)
{
    // don't bother verifying channel - we're about to die anyway
    // don't bother requiring empty packet - we're about to die anyway
    printf("connection: pty failure\r\n");
    ssh_kill(ss);
    return ss->c.state;
}


static connection_state_t 
connection_handle_shell_success(ssh_session_t *ss, Buffer *packet)
{
    uint32_t channel = buffer_get_int(packet);
    if (channel != ss->c.localChannel) {
        printf("connection: shell success on wrong channel (wanted %d, got %d)\r\n", 
               ss->c.localChannel, channel);
        ssh_kill(ss);
        return ss->c.state;
    }

    buffer_require_empty(packet);

    // nothing to send
    // fixme wire up tty if needed

    printf("Connected to host '%s'.\r\n", ss->s.hostname);

    return CONNECTION_STATE_RUNNING;
}


static connection_state_t 
connection_handle_shell_failure(ssh_session_t *ss, Buffer *packet)
{
    // don't bother verifying channel - we're about to die anyway
    // don't bother requiring empty packet - we're about to die anyway
    printf("connection: shell failure\r\n");
    ssh_kill(ss);
    return ss->c.state;
}


static connection_state_t
connection_handle_window_adjust(ssh_session_t *ss, Buffer *packet)
{
    uint32_t channel;
    uint32_t delta;
    
    channel = buffer_get_int(packet);
    if (channel != ss->c.localChannel) {
        printf("connection: window adjust got wrong channel (wanted %d, got %d)\r\n", ss->c.localChannel, channel);
        ssh_kill(ss);
        return ss->c.state;
    }

    delta = buffer_get_int(packet);
    buffer_require_empty(packet);
    ss->c.sendWindow += delta;
    return ss->c.state;
}


static connection_state_t
connection_handle_channel_close(ssh_session_t *ss, Buffer *packet)
{
    uint32_t channel;
    channel = buffer_get_int(packet);
    if (channel != ss->c.localChannel) {
        printf("connection: close got wrong channel (wanted %d, got %d)\r\n", ss->c.localChannel, channel);
        ssh_kill(ss);
        return ss->c.state;
    }
    buffer_require_empty(packet);

    ss->c.remoteCloseReceived = 1;
    ssh_close(ss);
    return ss->c.state;
}


static connection_state_t
connection_handle_data(ssh_session_t *ss, Buffer *packet, int closing)
{
    uint32_t channel;
    uint8_t *data;
    uint16_t len;

    connection_log("conn handle data: 0x%x (len 0x%x)", packet, packet->alloc);

    channel = buffer_get_int(packet);
    if (channel != ss->c.localChannel) {
        printf("connection: data got wrong channel (wanted %d, got %d)\r\n", ss->c.localChannel, channel);
        ssh_kill(ss);
        return ss->c.state;
    }
    data = buffer_get_string(packet, &len);
    buffer_require_empty(packet);
    connection_handle_typed_data(ss, 0, data, len, closing);
    arena_free(data);
    return ss->c.state;
}


static connection_state_t 
connection_handle_extended_data(ssh_session_t *ss, Buffer *packet, int closing)
{
    uint32_t channel;
    uint32_t type;
    uint8_t *data;
    uint16_t len;

    channel = buffer_get_int(packet);
    if (channel != ss->c.localChannel) {
        printf("connection: extended data got wrong channel (wanted %d, got %d)\r\n", ss->c.localChannel, channel);
        ssh_kill(ss);
        return ss->c.state;
    }
    type = buffer_get_int(packet);
    data = buffer_get_string(packet, &len);
    buffer_require_empty(packet);
    connection_handle_typed_data(ss, type, data, len, closing);
    arena_free(data);
    return ss->c.state;
}








/***********************************************************************
 * state handlers
 */


// fixme move to transport
static connection_state_t 
connection_state_waiting_for_userauth_service(ssh_session_t *ss, 
                                              uint8_t msg, 
                                              Buffer *packet)
{
    if (msg == SSH2_MSG_SERVICE_ACCEPT) {
        return connection_handle_userauth_service_accept(ss, packet);
    } else {
        return connection_handle_unexpected_message(ss, msg, packet);
    }
}


static connection_state_t 
connection_state_waiting_for_password(ssh_session_t *ss, 
                                      uint8_t msg, Buffer *packet)
{
    // not expecting anything in particular - GUI triggers state change
    if (msg == SSH2_MSG_USERAUTH_BANNER) {
        return connection_handle_banner(ss, packet);
    } else {
        return connection_handle_unexpected_message(ss, msg, packet);
    }
}


static connection_state_t
connection_state_waiting_for_userauth(ssh_session_t *ss, 
                                      uint8_t msg, Buffer *packet)
{
    if (msg == SSH2_MSG_USERAUTH_SUCCESS) {
        return connection_handle_userauth_success(ss, packet);
    } else if (msg == SSH2_MSG_USERAUTH_FAILURE) {
        return connection_handle_userauth_failure(ss, packet);
    } else if (msg == SSH2_MSG_USERAUTH_PASSWD_CHANGEREQ) {
        return connection_handle_password_change(ss, packet);
    } else if (msg == SSH2_MSG_USERAUTH_BANNER) {
        return connection_handle_banner(ss, packet);
    } else {
        return connection_handle_unexpected_message(ss, msg, packet);
    }
}


static connection_state_t
connection_state_waiting_for_channel_open(ssh_session_t *ss, 
                                          uint8_t msg, Buffer *packet)
{
    if (msg == SSH2_MSG_CHANNEL_OPEN_CONFIRMATION) {
        return connection_handle_channel_open(ss, packet);
    } else if (msg == SSH2_MSG_CHANNEL_OPEN_FAILURE) {
        return connection_handle_channel_open_failure(ss, packet);
    } else {
        return connection_handle_unexpected_message(ss, msg, packet);
    }
}


static connection_state_t 
connection_state_waiting_for_pty(ssh_session_t *ss, 
                                 uint8_t msg, Buffer *packet)
{
    if (msg == SSH2_MSG_CHANNEL_SUCCESS) {
        return connection_handle_pty_success(ss, packet);
    } else if (msg == SSH2_MSG_CHANNEL_FAILURE) {
        return connection_handle_pty_failure(ss, packet);
    } else if (msg == SSH2_MSG_CHANNEL_WINDOW_ADJUST) {
        return connection_handle_window_adjust(ss, packet);
    } else if (msg == SSH2_MSG_CHANNEL_CLOSE) {
        return connection_handle_channel_close(ss, packet);
    } else {
        return connection_handle_unexpected_message(ss, msg, packet);
    }
}


static connection_state_t 
connection_state_waiting_for_shell(ssh_session_t *ss, 
                                   uint8_t msg, Buffer *packet)
{
    if (msg == SSH2_MSG_CHANNEL_SUCCESS) {
        return connection_handle_shell_success(ss, packet);
    } else if (msg == SSH2_MSG_CHANNEL_FAILURE) {
        return connection_handle_shell_failure(ss, packet);
    } else if (msg == SSH2_MSG_CHANNEL_WINDOW_ADJUST) {
        return connection_handle_window_adjust(ss, packet);
    } else if (msg == SSH2_MSG_CHANNEL_CLOSE) {
        return connection_handle_channel_close(ss, packet);
    } else {
        return connection_handle_unexpected_message(ss, msg, packet);
    }
}


static connection_state_t 
connection_state_running(ssh_session_t *ss, 
                         uint8_t msg, Buffer *packet)
{
    if (msg == SSH2_MSG_CHANNEL_DATA) {
        return connection_handle_data(ss, packet, 0);
    } else if (msg == SSH2_MSG_CHANNEL_EXTENDED_DATA) {
        return connection_handle_extended_data(ss, packet, 0);
    } else if (msg == SSH2_MSG_CHANNEL_WINDOW_ADJUST) {
        return connection_handle_window_adjust(ss, packet);
    } else if (msg == SSH2_MSG_CHANNEL_CLOSE) {
        return connection_handle_channel_close(ss, packet);
    } else if (msg == SSH2_MSG_CHANNEL_EOF) {
        // silently ignore (fixme correct?)
        return ss->c.state;
    } else {
        return connection_handle_unexpected_message(ss, msg, packet);
    }
}



static connection_state_t 
connection_state_closing(ssh_session_t *ss, 
                         uint8_t msg, Buffer *packet)
{
    if (msg == SSH2_MSG_CHANNEL_DATA) {
        return connection_handle_data(ss, packet, 1);
    } else if (msg == SSH2_MSG_CHANNEL_EXTENDED_DATA) {
        return connection_handle_extended_data(ss, packet, 1);
    } else if (msg == SSH2_MSG_CHANNEL_WINDOW_ADJUST) {
        // ignore window adjust - we're sending no more data
        return ss->c.state;
    } else if (msg == SSH2_MSG_CHANNEL_CLOSE) {
        return connection_handle_channel_close(ss, packet);
    } else {
        return connection_handle_unexpected_message(ss, msg, packet);
    }
}


static connection_state_t 
connection_state_closed(ssh_session_t *ss, 
                        uint8_t msg, Buffer *packet)
{
    // shouldn't happen
    printf("connection: packet sent to connection_state_closed!\r\n");
    return connection_handle_unexpected_message(ss, msg, packet);
}

/***********************************************************************
 * incoming packet handler 
 */


// msg is gone from packet
void connection_receive_packet(ssh_session_t *ss, uint8_t msg, Buffer *packet)
{
    connection_state_t s = ss->c.state;

    connection_log("conn (state %d): received msg %d\r\n", ss->c.state, msg);

    switch (ss->c.state) {
    case CONNECTION_STATE_WAITING_FOR_USERAUTH_SERVICE:
        s = connection_state_waiting_for_userauth_service(ss, msg, packet);
        break;
    case CONNECTION_STATE_WAITING_FOR_PASSWORD:
    case CONNECTION_STATE_WAITING_FOR_NEW_PASSWORD:
        s = connection_state_waiting_for_password(ss, msg, packet);
        break;
    case CONNECTION_STATE_WAITING_FOR_USERAUTH:
        s = connection_state_waiting_for_userauth(ss, msg, packet);
        break;
    case CONNECTION_STATE_WAITING_FOR_CHANNEL_OPEN:
        s = connection_state_waiting_for_channel_open(ss, msg, packet);
        break;
    case CONNECTION_STATE_WAITING_FOR_PTY:
        s = connection_state_waiting_for_pty(ss, msg, packet);
        break;
    case CONNECTION_STATE_WAITING_FOR_SHELL:
        s = connection_state_waiting_for_shell(ss, msg, packet);
        break;
    case CONNECTION_STATE_RUNNING:
        s = connection_state_running(ss, msg, packet);
        break;
    case CONNECTION_STATE_CLOSING:
        s = connection_state_closing(ss, msg, packet);
        break;
    case CONNECTION_STATE_CLOSED:
        s = connection_state_closed(ss, msg, packet);
        break;

    default:
        printf("connection: BUSTED\r\n");
        break;
    }

    ss->c.state = s;
}


void connection_send_data(ssh_session_t *ss, uint8_t *bytes, uint16_t len)
{
    if (ss->c.state == CONNECTION_STATE_RUNNING) {
        // fixme enforce packet sizes and send window
        connection_send_channel_data(ss, ss->c.remoteChannel, bytes, len);
    }
}


/***********************************************************************
 * startup and shutdown 
 */

void connection_start(ssh_session_t *ss)
{
    if (ss->c.state == CONNECTION_STATE_STARTING) {
        printf("Logging in to host '%s'\r\n", ss->s.hostname);
        connection_send_service_request(ss, "ssh-userauth");
        ss->c.state = CONNECTION_STATE_WAITING_FOR_USERAUTH_SERVICE;
    }
}


static void connection_cleanup(ssh_session_t *ss)
{
    ss->c.state = CONNECTION_STATE_CLOSED;
}


void connection_kill(ssh_session_t *ss)
{
    if (ss->c.state != CONNECTION_STATE_CLOSED) {
        connection_cleanup(ss);
    }
}


// clean shutdown
// send CHANNEL_CLOSE and wait for remote, if necessary
void connection_close(ssh_session_t *ss)
{
    if (ss->c.state != CONNECTION_STATE_CLOSED) {
        if (! ss->c.channelOpen) {
            // channel not open - no two-way handshake needed
            connection_kill(ss);
        } else {
            // handle two-way channel close
            if (!ss->c.localCloseSent) {
                // tell remote that we want to close
                connection_send_channel_close(ss, ss->c.remoteChannel);
                ss->c.localCloseSent = 1;
            }
            if (!ss->c.remoteCloseReceived) {
                // continue processing packets until remote's channel close arrives
                ss->c.state = CONNECTION_STATE_CLOSING;
            } 
            if (ss->c.localCloseSent  &&  ss->c.remoteCloseReceived) {
                // two-way handshake done - really close
                ss->c.channelOpen = 0;
                ss->c.localCloseSent = 0;
                ss->c.remoteCloseReceived = 0;
                connection_kill(ss);
            }
        }
    }
}


int connection_is_open(ssh_session_t *ss) {
    return (ss->c.state != CONNECTION_STATE_STARTING  &&  
            ss->c.state != CONNECTION_STATE_CLOSING  &&  
            ss->c.state != CONNECTION_STATE_CLOSED);
}


int connection_is_closing(ssh_session_t *ss) {
    return (ss->c.state == CONNECTION_STATE_CLOSING);
}
