/**********
 * Copyright (c) 2003-2004 Greg Parker.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY GREG PARKER ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **********/

#include "includes.h"
#include "vt100.h"
#include "rsrc/rsrc.h"

#include "putty.h"

display *gdd = NULL;

static int inited = 0;

// bounds is lo-res
display *vt100_new(struct ssh_session_t *ss, RectangleType bounds) 
{
    if (!inited) {
        // first-time init
        init_ucs_tables();
        inited = 1;
    }


    gdd = arena_malloc(sizeof(display));


    // fixme hardcoded font
    gdd->font = (FontID)NanoFontID;
    gdd->font_width = 4;
    gdd->font_height = 6;

    // set some cfg options
    cfg.no_mouse_rep = 1; // xterm mouse doesn't work
    cfg.no_remote_resize = 1; // remote window manipulation doesn't work
    cfg.no_remote_charset = 1; // other fonts don't work 
    cfg.wrap_mode = 1; // wrap to next line

    term_init();
    vt100_set_bounds(gdd, bounds, NULL, NULL);

    return gdd;
}

void vt100_free(display *disp) 
{
    term_clrsb();
    arena_free(disp);
    gdd = NULL;
}

void vt100_write(display *disp, char *bytes, long len) 
{
    if (len > 0) {
        from_backend(0, bytes, len);
        term_out();
    }
}

void vt100_update(display *disp)
{
    term_update();
}

void vt100_scroll(display *disp, int lines) 
{
    term_scroll(0, lines);
}

void vt100_activate(display *disp) 
{
    // fixme?
    term_invalidate();
    term_update();
}

void vt100_deactivate(display *disp) 
{
    // fixme (test with server kill during password dialog)
}

void vt100_set_bounds(display *disp, RectangleType bounds, 
                      int *charsWide, int *charsHigh) 
{
    int cw, ch;

    // scale bounds to hi-res
    gdd->bounds = bounds;
    WinPushDrawState();
    WinSetCoordinateSystem(kCoordinatesDouble);
    WinScaleRectangle(&gdd->bounds); // convert from standard to current
    WinPopDrawState();

    cw = gdd->bounds.extent.x / gdd->font_width;
    ch = gdd->bounds.extent.y / gdd->font_height;

    term_size(ch, cw, 500/*scrollback lines fixme*/);     

    term_invalidate();
    term_update();

    if (charsWide) *charsWide = cw;
    if (charsHigh) *charsHigh = ch;
}


void vt100_size(display *disp, int *charsWide, int *charsHigh)
{
    if (charsWide) *charsWide = gdd->bounds.extent.x / gdd->font_width;
    if (charsHigh) *charsHigh = gdd->bounds.extent.y / gdd->font_height;
}


// x and y are LO-RES
// clickCount is 1 or more for mouse down, 0 for mouse up
void vt100_click(display *disp, int x, int y, int clickCount)
{
    Coord hix, hiy;
    int charx, chary;
    Mouse_Button button;
    Mouse_Action action;

    // scale point to hi-res
    WinPushDrawState();
    WinSetCoordinateSystem(kCoordinatesDouble);
    hix = WinScaleCoord(x, false); // convert from standard to current
    hiy = WinScaleCoord(y, false); // convert from standard to current
    WinPopDrawState();

    // calculate coordinate
    charx = (hix - gdd->bounds.topLeft.x) / gdd->font_width;
    chary = (hiy - gdd->bounds.topLeft.y) / gdd->font_height;

    // choose mouse button and mouse action
    button = MBT_LEFT;
    switch (clickCount) {
    case -1: action = MA_DRAG; break;
    case 0:  action = MA_RELEASE; break; // mouse up
    case 1:  action = MA_CLICK;   break; // mouse down
    case 2:  action = MA_2CLK;    break; // double-click
    case 3:  action = MA_3CLK;    break; // triple-click
    default: return; // ignore 4+ clicks (fixme?)
    }
    
    // fixme shift/ctrl/alt modifiers?
    // send at least one drag before release
    term_mouse(button, action, charx, chary, 0, 0, 0);
}


void vt100_deselect(display *disp)
{
    term_deselect();
}


void vt100_copy(display *disp)
{
    term_copy();
}

void vt100_paste(display *disp) 
{
    term_do_paste();
}

int vt100_selection_exists(display *disp) 
{
    return term_selection_exists();
}
